/*
Build: gcc -o fpfail.exe fpfail.c
*/

#include <stdio.h>
#include <stdlib.h>

const unsigned int N = 5;                               //vector length

double SumUpFPU(double *p,unsigned int m);              //protos
double SumUpXMM(double *p,unsigned int m);
                               
int main(int argc, char **argv)
{
  double V1[5] = {1e20,17.0,-10.0,130.0,-1e20};
  double V2[5] = {1e20,-10.0,130.0,-1e20,17.0};
  double V3[5] = {1e20,17.0,-1e20,-10.0,130.0};
  double V4[5] = {1e20,-10.0,-1e20,130.0,17.0};
  double V5[5] = {1e20,-1e20,17.0,-10.0,130.0};
  double V6[5] = {1e20,17.0,130.0,-1e20,-10.0};
  
  double sum1FPU,sum1XMM;
  double sum2FPU,sum2XMM;
  double sum3FPU,sum3XMM;
  double sum4FPU,sum4XMM;
  double sum5FPU,sum5XMM;
  double sum6FPU,sum6XMM;

  sum1FPU = SumUpFPU(&V1[0],N);
  sum1XMM = SumUpXMM(&V1[0],N);
  sum2FPU = SumUpFPU(&V2[0],N);
  sum2XMM = SumUpXMM(&V2[0],N);
  sum3FPU = SumUpFPU(&V3[0],N);
  sum3XMM = SumUpXMM(&V3[0],N);
  sum4FPU = SumUpFPU(&V4[0],N);
  sum4XMM = SumUpXMM(&V4[0],N);
  sum5FPU = SumUpFPU(&V5[0],N);
  sum5XMM = SumUpXMM(&V5[0],N);
  sum6FPU = SumUpFPU(&V6[0],N);
  sum6XMM = SumUpXMM(&V6[0],N);

  
  printf("Sum 1 (FPU) = %.2lf Sum 1 (XMM) = %.2f\n",sum1FPU,sum1XMM);
  printf("Sum 2 (FPU) = %.2lf Sum 2 (XMM) = %.2f\n",sum2FPU,sum2XMM);
  printf("Sum 3 (FPU) = %.2lf Sum 3 (XMM) = %.2f\n",sum3FPU,sum3XMM);
  printf("Sum 4 (FPU) = %.2lf Sum 4 (XMM) = %.2f\n",sum4FPU,sum4XMM);
  printf("Sum 5 (FPU) = %.2lf Sum 5 (XMM) = %.2f\n",sum5FPU,sum5XMM);
  printf("Sum 6 (FPU) = %.2lf Sum 6 (XMM) = %.2f\n",sum6FPU,sum6XMM);
  printf("\nRight Sum   = 137.00\n");

  return 0;
}

/*
SumUpFPU
Purpose:        Sum up a double vector with the FPU.
Input:          [ebp+8]  = pointer to vector
                [ebp+12] = vector length
                [ebp-8]  = function result
Output:         vector sum (right or wrong)                
*/
double SumUpFPU(double *p,unsigned int m)
{
  double s;
  __asm__ __volatile__(
    ".intel_syntax noprefix 			\n\t"
    "mov        edx,[ebp+8]                     \n\t"   //edx -> vector
    "mov        ecx,[ebp+12]                    \n\t"   //ecx = m (counter)
    "fldz                                       \n\t"   //sum = 0
    "1:                                         \n\t"
    "fadd       qword ptr [edx]                 \n\t"   //sum up
    "lea        edx,[edx+8]                     \n\t"   //update pointer
    "dec        ecx                             \n\t"
    "jnz        1b                              \n\t"
    "fstp       qword ptr [ebp-8]               \n\t"   //store result
    ".att_syntax prefix 			\n\t"
    :
    :"g"(p),"g"(m)
    :"%ecx","%edx"
    );
  return s;
}

/*
SumUpXMM
Purpose:        Sum up a double vector with XMM registers.
Input:          [ebp+8]  = pointer to vector
                [ebp+12] = vector length
                [ebp-8]  = function result
Output:         vector sum (right or wrong)                
*/
double SumUpXMM(double *p,unsigned int m)
{
  double s;
  __asm__ __volatile__(
    ".intel_syntax noprefix 			\n\t"
    "mov        edx,[ebp+8]                     \n\t"   //edx -> vector
    "mov        ecx,[ebp+12]                    \n\t"   //ecx = m (counter)
    "pxor       xmm0,xmm0                       \n\t"   //sum = 0
    "1:                                         \n\t"
    "addsd      xmm0,[edx]                      \n\t"   //sum up
    "lea        edx,[edx+8]                     \n\t"   //update pointer
    "dec        ecx                             \n\t"
    "jnz        1b                              \n\t"
    "movsd      [ebp-8],xmm0                    \n\t"   //store result
    ".att_syntax prefix 			\n\t"
    :
    :"g"(p),"g"(m)
    :"%ecx","%edx"
    );
  return s;
}

